package jp.ne.xuni.selene;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.rmi.Naming;
import java.rmi.RMISecurityManager;
import java.util.ArrayList;

import jp.ne.xuni.selene.views.TimeCounter;
import jp.ne.xuni.seleneserver.ISeleneServer;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;

public class SeleneClient {
	public static final String DEFAULT_HOST = "localhost";
	public static final String DEFAULT_REPOSITORY = "selene";

	private ISeleneServer obj;

	private String host = DEFAULT_HOST;
	private String repository = DEFAULT_REPOSITORY;

	public class Search extends Job {
		private static final int MAX_LINES = 50;

		private String query;
		private int cursorPosition = 0;
		private double attCoeff = 1.0;
		private double threshold = 0.5;
		private SearchResult[] result;

		public Search(String name) {
			super(name);
		}

		public void setQuery(String query) {
			this.query = query;
		}
		public void setCursorPosition(int cursorPosition) {
			this.cursorPosition = cursorPosition;
		}
		public void setAttCoeff(double attCoeff) {
			this.attCoeff = attCoeff;
		}
		public void setThreshold(double threshold) {
			this.threshold = threshold;
		}

		public SearchResult[] getSearchResult() {
			return result;
		}

		protected IStatus run(IProgressMonitor monitor) {
			if (obj == null && !lookup()) {
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Server \"" + host + "\" not found");
			} else if (query == null || query.length() == 0) {
				return new Status(IStatus.INFO, SelenePlugin.PLUGIN_ID,
					"There is nothing to search");
			}
			monitor.beginTask("Searching documents", 100);
			monitor.worked(10);

			try {
				BufferedReader in =
					new BufferedReader(new StringReader(obj.search(repository,
					query, cursorPosition, attCoeff, threshold)));
				monitor.worked(80);

				ArrayList<SearchResult> lines = parseResult(in);
				result = new SearchResult[lines.size()];
				result = lines.toArray(result);
			} catch (Exception e) {
				obj = null;
				System.out.println("SeleneClient error: " + e.getMessage());
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Search failed", e);
			}

			monitor.done();
			return new Status(IStatus.OK, SelenePlugin.PLUGIN_ID,
				"Search completed");
		}

		private ArrayList<SearchResult> parseResult(BufferedReader in)
			throws IOException {
			String line;
			int index;
			ArrayList<SearchResult> lines =
				new ArrayList<SearchResult>(MAX_LINES);
			while ((line = in.readLine()) != null) {
				if ((index = line.indexOf(ISeleneServer.SEPARATOR)) == -1) {
					continue;
				}
				lines.add(new DirectSearchResult(
					Double.parseDouble(line.substring(0, index)),
					host, repository, line.substring(index + 1)));
			}
			lines.trimToSize();
			return lines;
		}
	}

	public class FetchDocument extends Job {
		private SearchResult query;
		private FetchedDocument result;

		public FetchDocument(String name) {
			super(name);
		}

		public void setQuery(SearchResult query) {
			this.query = query;
		}

		public FetchedDocument getFetchResult() {
			return result;
		}

		protected IStatus run(IProgressMonitor monitor) {
			if (obj == null) {
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Server \"" + host + "\" not found");
			} else if (query == null) {
				return new Status(IStatus.INFO, SelenePlugin.PLUGIN_ID,
					"There is nothing to fetch");
			}
			monitor.beginTask("Fetching the document", 100);
			monitor.worked(10);

			try {
				result = new FetchedDocument(
					obj.getDocument(repository, query.getName()), query);
				monitor.worked(80);
			} catch (Exception e) {
				System.out.println("SeleneClient error: " + e.getMessage());
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Fetch failed", e);
			}

			monitor.done();
			return new Status(IStatus.OK, SelenePlugin.PLUGIN_ID,
				"Fetch completed");
		}
	}

	public class SearchSnippet extends Job {
		protected SearchResult query;
		private SearchResult[] original;
		protected Snippet snippet;
		protected int id = 0;
		protected TimeCounter tc;

		public SearchSnippet(String name) {
			super(name);
		}
		public SearchSnippet(String name, int id) {
			this(name);
			this.id = id;
		}
		public SearchSnippet(String name, int id, TimeCounter tc) {
			this(name, id);
			this.tc = tc;
		}

		public void setQuery(SearchResult query) {
			this.query = query;
		}
		public void setOriginal(SearchResult[] original) {
			this.original = original;
		}

		public Snippet getSnippet() {
			return snippet;
		}
		public int getId() {
			return id;
		}

		protected IStatus run(IProgressMonitor monitor) {
			FetchDocument fetchJob =
				new FetchDocument("Selene Snippet Search - fetch");
			fetchJob.setQuery(query);
			fetchJob.schedule();
			try {
				fetchJob.join();
			} catch (InterruptedException e) {
				System.out.println("SeleneClient error: " + e.getMessage());
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Snippet fetch failed", e);
			}
			if (!fetchJob.getResult().isOK()) {
				return new Status(IStatus.INFO, SelenePlugin.PLUGIN_ID,
					"Snippet fetch failed");
			}
			snippet = new Snippet(fetchJob.getFetchResult());

			if (tc != null) tc.recordTime(id); // for time recording
			Search searchJob = new Search("Selene Snippet Search - search");
			int bestSim = 0;
			int bestPosition = 0;
			while (snippet.hasMoreElements()) {
				int currentPosition = snippet.getCurrentLine();
				searchJob.setQuery(snippet.nextElement());
				searchJob.schedule();
				try {
					searchJob.join();
				} catch (InterruptedException e) {
					System.out.println(
						"SeleneClient error: " + e.getMessage());
					return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
						"Snippet search failed", e);
				}
				int currentSim = sim(original, searchJob.getSearchResult());
				if (bestSim <= currentSim) {
					bestSim = currentSim;
					bestPosition = currentPosition;
				}
			}
			snippet.setCurrentLine(bestPosition);
			if (tc != null) tc.recordTime(id); // for time recording

			return new Status(IStatus.OK, SelenePlugin.PLUGIN_ID,
				"Snippet search completed");
		}

		private int sim(SearchResult[] original, SearchResult[] snippet) {
			int counter = 0;
			for (int i = 0; i < original.length; i++) {
				for (int j = 0; j < snippet.length; j++) {
					if (original[i].getName().equals(snippet[j].getName())) {
						counter++;
					}
				}
			}
			return counter;
		}
	}

	public class FetchSnippet extends SearchSnippet {
		public FetchSnippet(String name) {
			super(name);
		}
		public FetchSnippet(String name, int id) {
			super(name, id);
		}
		public FetchSnippet(String name, int id, TimeCounter tc) {
			super(name, id, tc);
		}

		protected IStatus run(IProgressMonitor monitor) {
			FetchDocument fetchJob =
				new FetchDocument("Selene Snippet Search - fetch");
			fetchJob.setQuery(query);
			fetchJob.schedule();
			try {
				fetchJob.join();
			} catch (InterruptedException e) {
				System.out.println("SeleneClient error: " + e.getMessage());
				return new Status(IStatus.ERROR, SelenePlugin.PLUGIN_ID,
					"Snippet fetch failed", e);
			}
			if (!fetchJob.getResult().isOK()) {
				return new Status(IStatus.INFO, SelenePlugin.PLUGIN_ID,
					"Snippet fetch failed");
			}
			snippet = new Snippet(fetchJob.getFetchResult());

			if (tc != null) tc.recordTime(id); // for time recording
			snippet.setCurrentLine(query instanceof DirectSearchResult ?
				((DirectSearchResult)query).getLine() : 0);
			if (tc != null) tc.recordTime(id); // for time recording

			return new Status(IStatus.OK, SelenePlugin.PLUGIN_ID,
				"Snippet search completed");
		}
	}

	public SeleneClient() {
		super();
		if (System.getSecurityManager() == null) {
			System.setSecurityManager(new RMISecurityManager());
		}
	}

	public boolean lookup() {
		try {
			obj = (ISeleneServer)Naming.lookup("//" + host + "/SeleneServer");
			return true;
		} catch (Exception e) {
			obj = null;
			return false;
		}
	}

	public boolean setHost(String host) {
		this.host = host;
		return lookup();
	}
	public void setRepository(String repository) {
		this.repository = repository;
	}

	public String getHost() {
		return host;
	}
	public String getRepository() {
		return repository;
	}
}
