/* $Id: selene.c,v 1.3 2010/04/14 06:54:54 sodium Exp $ */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <geta/dtypes.h>
#include <geta/dw.h>
#include <geta/dwut.h>
#include <geta/nlog.h>
#include <geta/dxr.h>
#include <geta/dwsh.h>
#include <geta/dwam.h>
#include <geta/dwwt.h>

typedef struct syminfo syminfo;

#define MAX_LINES 50
#define SEPARATOR ','

#define CONTINUOUS_MODE 0x01

#define WAM_INIT_FAILED 1
#define HANDLE_NOT_SPECIFIED 2
#define HANDLE_NOT_FOUND 3
#define EXT_FILTER_NOT_FOUND 4
#define BUFFER_ALLOCATION_FAILED 5
#define STDIN_READ_FAILED 6
#define NO_INPUT 7
#define QUERY_ANALYSIS_FAILED 8
#define NO_MATCH 9
#define INVALID_PARAMETERS 10

#define ROUND(x, y) ((x) < floor(x) + (y) ? floor(x) : ceil(x))

/* static char GETAROOT[] = "/usr/local/geta"; */
#ifndef EXT_FILTER
#define EXT_FILTER "filter"
#endif
static char ext_filter[] = EXT_FILTER;

syminfo *wwstem(const char *text, WAM *w, int dir, const char *path,
	ssize_t *ndp, int cursor_position, double att_coeff, double threshold);
syminfo *wwstemv(const char *text, WAM *w, int dir, const char *path,
	char * const argv[], ssize_t *ndp, int cursor_position, double att_coeff,
	double threshold);
void makemk(char *marker, size_t size);
size_t writemk(int fd, const char *buf, size_t count, const char *marker);

int main(int argc, char **argv) {
	const char *handle;
/*	const char *ext_filter; */
	WAM *wam;
	char *buf;
	size_t size;
	int i;
	syminfo *q, *r;
	ssize_t qlen, rlen;
	int cursor_position = 0;
	double att_coeff = 1.0;
	double threshold = 0.5;
	char flags = 0;

/*	if (wam_init((char *)GETAROOT) == -1) { */
	if (wam_init(NULL) == -1) {
		return WAM_INIT_FAILED;
	}
	if (argc < 2) {
		return HANDLE_NOT_SPECIFIED;
	}
	handle = argv[1];
	if ((wam = wam_open((char *)handle)) == NULL) {
		return HANDLE_NOT_FOUND;
	}
/*
	if ((ext_filter = ci_value((char *)handle, "ext-filter")) == NULL) {
		return EXT_FILTER_NOT_FOUND;
	}
 */
	if (argc > 2 && !strcmp(argv[2], "-c")) {
		flags |= CONTINUOUS_MODE;
	}

	do {
		if (scanf("%zd%d%lf%lf%*1[\n]",
			&size, &cursor_position, &att_coeff, &threshold) < 1) {
			return INVALID_PARAMETERS;
		}
		if (size <= 0) {
			return NO_INPUT;
		}
		att_coeff = MIN(att_coeff, 1.0);
		threshold = MIN(fabs(threshold), 1.0);

		if ((buf = (char *)malloc(size + 1)) == NULL) {
			return BUFFER_ALLOCATION_FAILED;
		}
		if (fread(buf, 1, size, stdin) != size) {
			return STDIN_READ_FAILED;
		}
		*(buf + size) = '\0';
#ifdef PRINT_BUF
		fprintf(stderr, "%s%n", buf, &i);
		fprintf(stderr, "-- %d characters read --\n", i);
#endif

		if ((q = wwstem(buf, wam, WAM_COL, ext_filter, &qlen, cursor_position,
			att_coeff, threshold)) == NULL) {
			return QUERY_ANALYSIS_FAILED;
		}
		free(buf);
		if (qlen <= 0) goto no_output;

		rlen = MAX_LINES;
#ifdef RECORD_TIME
		fprintf(stderr, "%d ", (int)clock() * 1000 / CLOCKS_PER_SEC);
#endif
		r = wsh(q, qlen, wam, WAM_COL, WT_SMART, &rlen, NULL);
#ifdef RECORD_TIME
		fprintf(stderr, "%d\n", (int)clock() * 1000 / CLOCKS_PER_SEC);
#endif
		if (r == NULL) {
no_output:
			printf("0\n");
		} else {
			qsort(r, rlen, sizeof *r, sym_compar);

			printf("%zd\n", rlen);
			for (i = 0; i < rlen; i++) {
				printf("%f%c%s\n", r[i].weight,
					SEPARATOR, wam_id2name(wam, WAM_ROW, r[i].id));
			}
			free(r);
		}
		free(q);
		fflush(stdout);
		fflush(stderr);
	} while (flags & CONTINUOUS_MODE);

	wam_close(wam);
	return 0;
}

/* taken from wstem.c */
syminfo *wwstem(const char *text, WAM *w, int dir, const char *path,
	ssize_t *ndp, int cursor_position, double att_coeff, double threshold) {
	char *argv[3];

	argv[0] = (char *)path;
	argv[1] = "-1";
	argv[2] = NULL;
	return wwstemv(text, w, dir, path, argv, ndp, cursor_position, att_coeff,
		threshold);
}

syminfo *wwstemv(const char *text, WAM *w, int dir, const char *path,
	char * const argv[], ssize_t *ndp, int cursor_position, double att_coeff,
	double threshold) {
	struct itimerval it = {{0, 0}, {0, 0}}, ot;

	int status;
	pid_t cpid, wpid;
	int e;
/*
	wstem(writer) -> fds0[1] == fds0[0] =dup2=> 0 -> path (child)
	wstem(parent) <- fds1[0] == fds1[1] =dup2=> 1 <- path
 */
	int fds0[2];
	int fds1[2];
	FILE *f;

	size_t i, j, k;
	syminfo *p, *r;
	char *buf;
	size_t l, buflen;
	size_t rmax;

	int lnum = -1;
	char marker[] = "abcdefghijklmnopqrstuvwxyzabcdefghijklmnopqrstuvwxyz\n";

	r = NULL;
	buf = NULL;
	f = NULL;
	fds0[0] = fds0[1] = fds1[0] = fds1[1] = -1;

	l = strlen(text);
	rmax = l * 4;
	buflen = l + 513;

	makemk(marker, strlen(marker) - 1);

	if (!(buf = (char *)malloc(buflen))) {
		nperror("malloc buf");
		goto error;
	}
	if (!(r = (syminfo *)calloc(rmax, sizeof *r))) {
		nperror("calloc r");
		goto error;
	}
	if (pipe(fds0)==-1) {
		nperror("pipe");
		goto error;
	}
	if (pipe(fds1)==-1) {
		nperror("pipe");
		goto error;
	}

	switch (cpid = fork()) {
	case -1:
		goto error;
	case 0: /* child */
		setitimer(ITIMER_VIRTUAL, &it, &ot);

		close(fds1[0]), fds1[0] = -1;
		close(fds0[1]), fds0[1] = -1;
		if (fds0[0]!=0) {
			dup2(fds0[0], 0);
			close(fds0[0]), fds0[0] = -1;
		}
		if (fds1[1]!=1) {
			dup2(fds1[1], 1);
			close(fds1[1]), fds1[1] = -1;
		}
		execv(path, argv);
		nperror(path);
		_exit(1);
		break;
	default: /* parent */
		close(fds0[0]), fds0[0] = -1;
		close(fds1[1]), fds1[1] = -1;
		break;
	}

	switch (wpid = fork()) {
	case -1:
		goto error;
	case 0: /* writer */
		setitimer(ITIMER_VIRTUAL, &it, &ot);

		if (writemk(fds0[1], text, l, marker) != l) {
			goto error;
		}
		close(fds0[1]), fds0[1] = -1;
		close(fds1[0]), fds1[0] = -1;
		_exit(0);
		break;
	default: /* parent */
		close(fds0[1]), fds0[1] = -1;
		break;
	}

	if (!(f = fdopen(fds1[0], "r"))) {
		goto error;
	}
	fds1[0] = -1;

	for (i=0; fgets(buf, buflen, f);) {
		idx_t id;
		char *q;
		int dist;

		if ((q = index(buf, '\n')) != NULL) {
			*q = '\0';
		}
		if (!strncmp(buf, marker, strlen(marker) - 1)) {
			lnum++;
		} else if ((id = wam_name2id(w, dir, buf))) {
			if (i>=rmax) {
				goto error;
			}
			r[i].id = id;
			dist = abs(lnum - cursor_position);
			r[i++].weight = 1.0 / (1.0 + (dist - att_coeff * dist));
		}
	}
	fclose(f), f = NULL;

	qsort(r, i, sizeof *r, sym_icompar);

	for (j=0, p=r; j<i; j=k, p++) {
		for (k=j+1; r[j].id == r[k].id; k++) {
			r[j].weight += r[k].weight;
		}
		p->TF_d = ROUND(r[j].weight, threshold);
		if (!p->TF_d) {
			p--;
			continue;
		}
		p->id = r[j].id;
		p->TF = wam_freq_sum(w, dir, p->id);
		p->weight = 0.0;
		p->attr = WSH_OR;
	}
	i = p-r;

	if ((e = wait4(cpid, &status, 0, NULL))==-1) {
/*
		fprintf(stderr, "wait4 C=%d: %s\n", (int)cpid, strerror(errno));
 */
	}
	if ((e = wait4(wpid, &status, 0, NULL))==-1) {
/*
		fprintf(stderr, "wait4 W=%d: %s\n", (int)wpid, strerror(errno));
 */
	}

	*ndp = i;

cleanup:
	zfree(buf);
	if (fds0[0] != -1) {
		close(fds0[0]);
	}
	if (fds0[1] != -1) {
		close(fds0[1]);
	}
	if (fds1[0] != -1) {
		close(fds1[0]);
	}
	if (fds1[1] != -1) {
		close(fds1[1]);
	}
	if (f) {
		fclose(f);
	}

	return r;

error:
	zfree(r);
	r = NULL; /* important! */
	goto cleanup;
}
/* until here */

void makemk(char *marker, size_t size) {
	int i;

	srand(time(NULL));
	for (i = 0; i < size; i++) {
		marker[i] = 'a' + (int)(26.0 * rand() / (RAND_MAX + 1.0));
	}
}

size_t writemk(int fd, const char *buf, size_t count, const char *marker) {
	size_t size = 0, size_written = 0;

	while (size_written < count) {
		write(fd, marker, strlen(marker));
		while (buf[size] != '\n' && size < count) {
			size++;
		}
		size = MIN(size + 1, count);
		write(fd, buf + size_written, size - size_written);
		size_written = size;
	}
	return size_written;
}
